/*
 * gtkDPS font selection test program by Hideki FUJIMOTO
 * 
 */ 
#include <stdio.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include "gtkDPS.h"

static GtkWidget * new_fontsel_diag();
static void init_gui (int argc, char *argv[]);
static void print_fontsel(GtkButton * button_no_use, 
			  GtkDPSFontSelection * fontsel);

int
main (int argc, char *argv[])
{
  const gchar * version;
  gtk_set_locale();
  gtk_init (&argc, &argv);
  gtk_dps_init (&argc, &argv);
  version = gtk_dps_version();
  fprintf(stderr, "GtkDPS version %s\n", version);
  
  init_gui (argc, argv);
  gtk_main ();
  return 0;
}

static void
init_gui (int argc, char *argv[])
{
  GtkWidget *fontsel_diag;

  fontsel_diag = new_fontsel_diag(argv[1]?argv[1]:NULL);

  gtk_window_set_wmclass (GTK_WINDOW (fontsel_diag), "font selection dialog", "Font selection dialog");
  gtk_window_set_policy (GTK_WINDOW (fontsel_diag), TRUE, TRUE, FALSE);
  gtk_widget_set_name (fontsel_diag, "Font selection dialog");
  gtk_signal_connect (GTK_OBJECT (fontsel_diag), "destroy",
		      GTK_SIGNAL_FUNC (gtk_exit), NULL);
  gtk_signal_connect (GTK_OBJECT (fontsel_diag), "delete_event",
		      GTK_SIGNAL_FUNC (gtk_exit), NULL);
  gtk_widget_show(fontsel_diag);
}

static GtkWidget *
new_fontsel_diag(gchar * string)
{
  GtkWidget *fontsel_diag;
  GtkDPSFontSelection *fontsel;
  GtkButton * ok_button;
  GtkButton * cancel_button;
  GtkWidget * tmp;

  fontsel_diag = gtk_dps_font_selection_dialog_new ("Font selection dialog");
  
  tmp = GTK_DPS_FONT_SELECTION_DIALOG(fontsel_diag)->fontsel;
  fontsel = GTK_DPS_FONT_SELECTION(tmp);

  if (string)
    gtk_dps_font_selection_set_preview_text(fontsel, string);
  
  tmp = GTK_DPS_FONT_SELECTION_DIALOG(fontsel_diag)->ok_button;
  ok_button = GTK_BUTTON(tmp);
  gtk_signal_connect (GTK_OBJECT (ok_button), "clicked",
		      print_fontsel, fontsel);

  tmp = GTK_DPS_FONT_SELECTION_DIALOG(fontsel_diag)->cancel_button;
  cancel_button = GTK_BUTTON(tmp);
  gtk_signal_connect (GTK_OBJECT (cancel_button), "clicked",
		      GTK_SIGNAL_FUNC (gtk_exit), NULL);

  return fontsel_diag;
}

static void 
print_fontsel(GtkButton * button_no_use, GtkDPSFontSelection * fontsel)
{
  gchar * font_name ;
  gint font_size;
  gchar * preview_text;

  fprintf(stderr, "---\n");
  font_name = gtk_dps_font_selection_get_font_name(fontsel);
  fprintf(stderr, "Font name: %s\n", font_name);

  font_size = gtk_dps_font_selection_get_font_size(fontsel);
  fprintf(stderr, "Font size: %d\n", font_size);

  preview_text = gtk_dps_font_selection_get_preview_text(fontsel);
  fprintf(stderr, "Preview text: %s\n", preview_text);
}
